/* catchball2 .. SPE-to-SPE data transfer sample    |
|   2007.10.24 Naoki FUJIEDA, Arch Lab. TOKYO TECH */

#include "define.h"

#ifdef _PPESOURCE_

void initialize(spe_context_ptr_t *spe)
{
    int i, j;

    for (i = 0; i < SPE_NUM; i++) {
        dmapack[i].spe_num = SPE_NUM;
        dmapack[i].spe_id  = i + 1;
        dmapack[i].ppe2spe[0] = (uint) data;
        for (j = 0; j < SPE_NUM; j++)
            dmapack[i].ppe2spe[j + 1] = (uint) spe_ls_area_get(spe[j]);
        dmapack[i].ppe2spe[SPE_NUM + 1] = (uint) sync;
    }

    for (i = 0; i < DATA_NUM; i++) {
        data[i] = 1;
    }
    sync[0] = 1;
#ifdef MEMIMAGE
    int *p;
    FILE *fp;
    if ((fp = fopen(MEMIMAGE, "wb")) == NULL) {
        fprintf(stderr, "file can't open.\n");
        return;
    }
    fprintf(fp, "SimCell_PPE_MEMORY_IMAGE_FILE_VERSION_1.0\n");
    
    p = (int *)dmapack;
    for (i = 0; i < sizeof(dmapack) / 4; i++)
        fprintf(fp, "@mem %08x %08x\n", (uint) dmapack + 4 * i, p[i]);
    for (i = 0; i < DATA_NUM; i++)
        fprintf(fp, "@mem %08x %08x\n", (uint) data + 4 * i, data[i]);
    fprintf(fp, "@mem %08x %08x\n", (uint) sync, sync[0]);
    
    for (i = 0; i < SPE_NUM; i++)
        fprintf(fp, "@map %08x %08x\n", i + 1,
                (uint) spe_ls_area_get(spe[i]));
    
    for (i = 0; i < SPE_NUM; i++)
        fprintf(fp, "@mal %08x %08x\n", i + 1, (uint) &dmapack[i]);
    
    fprintf(fp, "@end %08x %08x\n", 0, 0);
    fclose(fp);
#endif /* MEMIMAGE */
}

void finalize()
{
    int i;
    printf("RESULT: ");
    for (i = 0; i < DATA_NUM; i++)
        printf("%d ", data[i]);
    puts("");
}

#endif /* _PPESOURCE_ */

/**********************************************************************/

#ifdef _SPESOURCE_

inline void wait_dma(int dma_tag)
{
    spu_writech(MFC_WrTagMask, 1 << dma_tag);
    spu_mfcstat(MFC_TAG_UPDATE_ALL);
}

void waitsync()
{
    int addr = dmapack.ppe2spe[dmapack.spe_num + 1];
    sync[0] = 0;
    printf("SPE %d wait for sync\n", dmapack.spe_id);
    while (sync[0] != dmapack.spe_id) {
        spu_mfcdma32((void *) sync, addr,
                     sizeof(sync), DMA_TAG0, MFC_GET_CMD);
        wait_dma(DMA_TAG0);
    }
}

void multiply()
{
    int i;

    printf("SPE %d get data.\n", dmapack.spe_id);
    int addr = dmapack.ppe2spe[dmapack.spe_id - 1];
    if (dmapack.spe_id != 1)
        addr += (int) data;

    spu_mfcdma32((void *) data, addr,
                 sizeof(data), DMA_TAG0, MFC_GET_CMD);
    wait_dma(DMA_TAG0);

    for (i = 0; i < DATA_NUM; i++)
        data[i] *= (i + 1);
}

void sendsync()
{
    int addr;
    if (dmapack.spe_id != dmapack.spe_num) {
        printf("SPE %d update sync\n", dmapack.spe_id);
        sync[0] = dmapack.spe_id + 1;
        addr = dmapack.ppe2spe[dmapack.spe_num + 1];
        spu_mfcdma32((void *) sync, addr,
                     sizeof(sync), DMA_TAG0, MFC_PUT_CMD);
        wait_dma(DMA_TAG0);
    } else {
        printf("SPE %d write result to memory\n", dmapack.spe_id);
        addr = dmapack.ppe2spe[0];
        spu_mfcdma32((void *) data, addr,
                     sizeof(data), DMA_TAG0, MFC_PUT_CMD);
        wait_dma(DMA_TAG0);
    }
    printf("SPE %d finish\n", dmapack.spe_id);
}

int main()
{
    spe_initialize();
    waitsync();
    multiply();
    sendsync();
    return 0;
}

#endif /* _SPESOURCE_ */
