/* mand .. Mandelbrot set program for Cell BE          |
|   2007.07.11 Naoki FUJIEDA, Arch Lab. TOKYO TECH    */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include "define.h"
#include "bmpgen.h"

rect rc;

int main(int argc, char **argv)
{
    int status;
    Pixel *pict_data;
    uint i, tmp, bright;
    struct timeval tv_end, tv_start;
    float a, b, x1, x2, y1, y2, r;
    uint xx, yy, j;

    // set domain
    rc.left   = -2.0;
    rc.top    = -1.25;
    rc.width  =  2.5;
    rc.height =  2.5;

    // memory allocate
    if (posix_memalign((void **)&pict_data, DMA_ALIGN, PIC_MEM)) {
        fprintf(stderr,"error: memalign failed\n");
        return(1);
    }
    
    gettimeofday(&tv_start, NULL);

    for (i = 0; i < PIC_NUM; i++) {
        xx = i % PIC_WIDTH;
        yy = i / PIC_WIDTH;
        a = rc.left + rc.width / PIC_WIDTH * xx;
        b = rc.top + rc.height / PIC_HEIGHT * yy;
        x1 = 0;
        y1 = 0;
        for (j = 1; j < MAX_CALC; j++) {
            x2 = x1 * x1 - y1 * y1 + a;
            y2 = x1 * y1 * 2 + b;
            r = x2 * x2 + y2 * y2;
            if (r > 4) break;
            x1 = x2;
            y1 = y2;
        }
        pict_data[i].color = j & (-(j != MAX_CALC));
    }

    gettimeofday(&tv_end, NULL);
    printf("Time = %d usec.\n",(int) ((tv_end.tv_usec - tv_start.tv_usec)
           + (tv_end.tv_sec - tv_start.tv_sec) * 1000000));

    // write result
    for (i = 0; i < PIC_NUM; i++) {
        if (pict_data[i].color != 0) {
            tmp = (pict_data[i].color - 1) % 28;
            bright = ((tmp / 7) * 50 + 105);
            tmp = (tmp % 7) + 1;
            pict_data[i].RGB.R = ((tmp & 0x4) != 0) * bright;
            pict_data[i].RGB.G = ((tmp & 0x2) != 0) * bright;
            pict_data[i].RGB.B = (tmp & 0x1) * bright;
            pict_data[i].RGB.reserved = 0;
        }
    }
    bmpgen("result.bmp", pict_data, PIC_WIDTH, PIC_HEIGHT);

    free(pict_data);
    return 0;
}
