/* bmpgen .. bi-endian .bmp file generater         |
|  2007.06.19 Naoki FUJIEDA, Arch Lab. TOKYO TECH */

#include "bmpgen.h"

void bmpgen_chHWord(char *file)
{
    char temp;

    temp = *file;
    *file = *(file + 1);
    *(file + 1) = temp;
}

void bmpgen_chWord(char *file)
{
    char temp;

    temp = *file;
    *file = *(file + 3);
    *(file + 3) = temp;
    temp = *(file + 1);
    *(file + 1) = *(file + 2);
    *(file + 2) = temp;
}

void bmpgen_writebfh(int fd, BITMAPFILEHEADER bfh)
{
    if (ENDIAN == BIG) {
        bmpgen_chHWord((char *) &bfh.bfType);
        bmpgen_chWord ((char *) &bfh.bfSize);
        bmpgen_chWord ((char *) &bfh.bfOffBits);
    }
    write(fd, &bfh, sizeof(BITMAPFILEHEADER));
}

void bmpgen_writebih(int fd, BITMAPINFOHEADER bih)
{
    if (ENDIAN == BIG) {
        bmpgen_chWord ((char *) &bih.biSize);
        bmpgen_chWord ((char *) &bih.biWidth);
        bmpgen_chWord ((char *) &bih.biHeight);
        bmpgen_chHWord((char *) &bih.biPlanes);
        bmpgen_chHWord((char *) &bih.biBitCount);
    }
    write(fd, &bih, sizeof(BITMAPINFOHEADER));
}

void bmpgen_check_endian()
{
    unsigned short bom = 0xfeff;
    unsigned char *checker;

    checker = (unsigned char *) &bom;
    ENDIAN = (*checker == 0xfe) ? BIG : LITTLE;
}

int bmpgen(char *filename, Pixel *image, int width, int height)
{
    BITMAPFILEHEADER bfh;
    BITMAPINFOHEADER bih;
    int fd;
    char *temp;
    unsigned int col;
    int i, imax, x, y;
    int linelength;
    int is8bit;
    Pixel white;
    Pixel palette[256];

    white.RGB.R = 255;
    white.RGB.G = 255;
    white.RGB.B = 255;

    // Check Endian (only when first use)
    if (ENDIAN == UNKNOWN)
        bmpgen_check_endian();

    // Check Invalid arg.
    if ((filename == NULL) || (image == NULL))
        return(1);
    if ((height <= 0) || (width <= 0))
        return(1);

    for (i = 0; i < 255; i++)
        palette[i].color = 0; 
    palette[255].color = white.color;

    // Count #Color
    imax = 0;
    for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++) {
            col = getpixel(image, x, y, width)->color;
            if (col != white.color)
                for (i = 0; i <= imax; i++)
                    if (col == palette[i].color)
                        break;
            if (i == imax + 1) {
                imax++;
                if (imax == 255) {
                    y = height;
                    break;
                }
                palette[i].color = col;
            }
        }
    }

    is8bit = (imax != 255);

    // allocate temporary data memory
    if (is8bit) {
        linelength = align_4b(width);
    } else {
        linelength = align_4b(width * 3);
    }
    temp = (char *) malloc(linelength);
    for (i = 0; i < linelength; i++)
        temp[i] = 0;
    
    // open file
    if ((fd = creat(filename, 0644)) == -1) {
        fprintf(stderr, "bmpgen: file open error\n");
        return(1);
    }

    // write bmp header
    bfh.bfType = BFTYPE_BM;
    bfh.bfReserved1 = 0;
    bfh.bfReserved2 = 0;
    bih.biSize = sizeof(BITMAPINFOHEADER);
    bih.biWidth = width;
    bih.biHeight = height;
    bih.biPlanes = 1;
    bih.biCompression = 0;
    bih.biSizeImage = 0;
    bih.biXPixPerMeter = 0;
    bih.biYPixPerMeter = 0;
    bih.biClrUsed = 0;
    bih.biClrImportant = 0;
    if (is8bit) {
        bfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) +
                        sizeof(palette);
        bfh.bfSize = bfh.bfOffBits + align_4b(width) * height;
        bih.biBitCount = 8;
    } else {
        bfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
        bfh.bfSize = bfh.bfOffBits + align_4b(width * 3) * height;
        bih.biBitCount = 24;
    }

    bmpgen_writebfh(fd, bfh);
    bmpgen_writebih(fd, bih);

    // write palette(256 colors)
    if (is8bit)
        write(fd, palette, sizeof(palette));

    // write bitmap data
    for (y = height - 1; y >= 0; y--) {
        for (x = 0; x < width; x++) {
            if (is8bit) {
                col = getpixel(image, x, y, width)->color;
                for (i = 0; i < 255; i++)
                    if (col == palette[i].color)
                        break;
                temp[x] = i;
            } else {
                temp[x * 3 + 0] = getpixel(image, x, y, width)->RGB.B;
                temp[x * 3 + 1] = getpixel(image, x, y, width)->RGB.G;
                temp[x * 3 + 2] = getpixel(image, x, y, width)->RGB.R;
            }
        }
        write(fd, temp, linelength);
    }

    // finalize
    close(fd);
    free(temp);
    return(0);
}
