/* Standard Library File for SimCell Regulation     |
|                  2007.11.01 Arch Lab. TOKYO TECH */

#define _GLOBAL_
#include "define.h"

#ifdef _PPESOURCE_

void initialize(spe_context_ptr_t *);
void finalize();

void *sperunfunc(void *arg)
{
    unsigned int entry = SPE_DEFAULT_ENTRY;
    spe_stop_info_t stop_info;
    int ret;
    ret = spe_context_run((spe_context_ptr_t) arg,
                          &entry, 0, NULL, NULL, &stop_info); 
    if (ret < 0)
        fprintf(stderr, "warning: an error occured in SPE thread\n");
    return NULL;
}

int main(int argc, char **argv)
{
    pthread_t spe_thread[SPE_NUM];
    spe_context_ptr_t spe[SPE_NUM];
    spe_program_handle_t *spe_program;
    int i;
    uint temp;

    /***** (Step 1) invoke spe threads *****/
    if ((spe_program = spe_image_open(SPE_FILE_NAME)) == NULL) {
        fprintf(stderr,"error: can't open %s\n", SPE_FILE_NAME);
        exit(1);
    }
    for (i = 0; i < SPE_NUM; i++) {
        if ((spe[i] = spe_context_create(0, NULL)) == NULL) {
            fprintf(stderr,"error: spe_context_create failed\n");
            exit(1);
        }
        if (spe_program_load(spe[i], spe_program)) {
            fprintf(stderr,"error: spe_program_load failed\n");
            exit(1);
        }
    }
    for (i = 0; i < SPE_NUM; i++) {
        if (pthread_create(&spe_thread[i], NULL, sperunfunc, spe[i])) {
            fprintf(stderr,"error: can't create SPE thread\n");
            exit(1);
        }
    }
    /***** (Step 2) initialize dmapack *****/
    initialize(spe);

    /***** (Step 3) send pointer of dmapack *****/
    for (i = 0; i < SPE_NUM; i++) {
        temp = (uint) &dmapack[i];
        spe_in_mbox_write(spe[i], &temp, 1, SPE_MBOX_ALL_BLOCKING);
    }

    /***** Wait SPE thread   *****/
    for (i = 0; i < SPE_NUM; i++)
        pthread_join(spe_thread[i], NULL);

    /***** User's Finalize and Close SPE *****/
    finalize();
    for (i = 0; i < SPE_NUM; i++)
        spe_context_destroy(spe[i]);
    spe_image_close(spe_program);

    return 0;
}

#endif /* _PPESOURCE_ */

/**********************************************************************/

#ifdef _SPESOURCE_

void spe_initialize()
{
    /***** (Step 4) receive the dmapack address *****/
    ea_dmapack = spu_read_in_mbox();

    /***** (Step 5) receive the dmapack contents *****/
    spu_mfcdma32((void *) &dmapack, ea_dmapack,
                 sizeof(dmapack_t), DMA_TAG0, MFC_GET_CMD);
    spu_writech(MFC_WrTagMask, 1 << DMA_TAG0);
    spu_mfcstat(MFC_TAG_UPDATE_ALL);
}

#endif /* _SPESOURCE_ */
