/* mand .. Mandelbrot set program for Cell BE       |
|   2007.10.24 Naoki FUJIEDA, Arch Lab. TOKYO TECH */

#include "define.h"

#ifdef _PPESOURCE_

#include <string.h>
#include <sys/time.h>

uint64 gettime()
{
    static uint64 start = 0;
    uint64 tm;
    struct timeval tv;

    gettimeofday(&tv, NULL);
    tm = tv.tv_sec * 1000000ull + tv.tv_usec;
    if (start == 0)
        start = tm;
    return tm - start;
}

void initialize(spe_context_ptr_t *spe)
{
    int i;
    float ftemp;

    for (i = 0; i < SPE_NUM; i++) {
        dmapack[i].spe_num = SPE_NUM;
        dmapack[i].spe_id  = i + 1;
        dmapack[i].ppe2spe[0] = (int) pict_data;
        ftemp = -2.0;  // rc.left
        dmapack[i].ppe2spe[1] = *(uint *) &ftemp;
        ftemp = -1.25; // rc.top
        dmapack[i].ppe2spe[2] = *(uint *) &ftemp;
        ftemp =  2.5;  // rc.width = rc.height
        dmapack[i].ppe2spe[3] = *(uint *) &ftemp;
        dmapack[i].ppe2spe[4] = dmapack[i].ppe2spe[3];
    }

    printf("** setting **\n");
    printf("%d SPEs, %s SIMD\n", SPE_NUM,
           strcmp(SPE_FILE_NAME, "./spe_simd") ? "without" : "with");
    printf("Size = %d x %d\n", PIC_WIDTH, PIC_HEIGHT);
    printf("Loop Limit = %d\n\n", MAX_CALC);
#ifdef MEMIMAGE
    int *p;
    FILE *fp;
    if ((fp = fopen(MEMIMAGE, "wb")) == NULL) {
        fprintf(stderr, "file can't open.\n");
        return;
    }
    fprintf(fp, "SimCell_PPE_MEMORY_IMAGE_FILE_VERSION_1.0\n");
    
    p = (int *)dmapack;
    for (i = 0; i < sizeof(dmapack) / 4; i++)
        fprintf(fp, "@mem %08x %08x\n", (uint) dmapack + 4 * i, p[i]);
    for (i = 0; i < SPE_NUM; i++)
        fprintf(fp, "@mal %08x %08x\n", i + 1, (int)&dmapack[i]);
    
    fprintf(fp, "@end %08x %08x\n", 0, 0);
    fclose(fp);
#endif /* MEMIMAGE */
    gettime();
}

void finalize()
{
    int i, tmp;
    printf("Processing time: %8.3f sec.\n", ((double) gettime()) / 1000000);

    // write result
    for (i = 0; i < PIC_NUM; i++) {
        if (pict_data[i].color != 0) {
            tmp = (pict_data[i].color - 1) % 30;
            pict_data[i].color = 0;
            if (tmp < 5) {
                pict_data[i].RGB.R = 255;
                pict_data[i].RGB.G = 51 * (tmp % 5);
            } else if (tmp < 10) {
                pict_data[i].RGB.R = 255 - 51 * (tmp % 5);
                pict_data[i].RGB.G = 255;
            } else if (tmp < 15) {
                pict_data[i].RGB.G = 255;
                pict_data[i].RGB.B = 51 * (tmp % 5);
            } else if (tmp < 20) {
                pict_data[i].RGB.G = 255 - 51 * (tmp % 5);
                pict_data[i].RGB.B = 255;
            } else if (tmp < 25) {
                pict_data[i].RGB.B = 255;
                pict_data[i].RGB.R = 51 * (tmp % 5);
            } else if (tmp < 30) {
                pict_data[i].RGB.B = 255 - 51 * (tmp % 5);
                pict_data[i].RGB.R = 255;
            }
        }
    }
    bmpgen("result.bmp",(Pixel *) pict_data, PIC_WIDTH, PIC_HEIGHT);
}

#endif /* _PPESOURCE_ */

/**********************************************************************/

#ifdef _SPESOURCE_

inline void wait_dma(uint tagid)
{
    spu_writech(MFC_WrTagMask, 1 << tagid);
    spu_mfcstat(MFC_TAG_UPDATE_ALL);
}

void large_dma(volatile void* ls, uint ea, uint size, uint tagid, uint cmd)
{
//    printf("DMA ls=%08x ea=%08x size=%05x tagid=%02d cmd=%02x\n",
//          (uint) ls, ea, size, tagid, cmd);
    uint dma_off, dma_size;
    dma_off = 0;
    while (size) {
        dma_size = (size < DMA_SIZE_MAX) ? size : DMA_SIZE_MAX;
        spu_mfcdma32(ls + dma_off, ea + dma_off, dma_size, tagid, cmd);
        size -= dma_size;
        dma_off += dma_size;
    }
}

void send_mem(uint num, uint block)
{
    large_dma(&mem[BLOCK_NUM / 2 * num],
              ea_pic + BLOCK_NUM * sizeof(uint) * block
              + BLOCK_NUM * sizeof(uint) / 2 * num,
              BLOCK_NUM * sizeof(uint) / 2, num, MFC_PUT_CMD);
    wait_dma(1 - num);
}

int calc()
{
    uint i, j;
    uint spe_num, spe_no;
    uint block;
    rect_t rc;
#ifdef SIMD
    // with SIMD
    v_float a, b, x1, x2, y1, y2, r;
    v_float xx, yy;
    v_float xstep, ystep;
    v_uint cnt, over, chk;
    v_uint *memp;
    uint temp[8];
#else
    // without SIMD
    float a, b, x1, x2, y1, y2, r;
    float xstep, ystep;
    uint xx, yy;
#endif

    // initialize
    mand = (volatile mand_t *) &dmapack;
    ea_pic = mand->ea_pic;
    rc = mand->rc;
    spe_num = mand->spe_num;
    spe_no = mand->spe_id - 1;

    // main loop
    for (block = spe_no; block < TOTAL_BLOCK; block += spe_num) {
#ifdef SIMD
        // with SIMD
        memp = (v_uint *) mem;
        xstep = spu_splats(rc.width / PIC_WIDTH);
        ystep = spu_splats(rc.height / PIC_HEIGHT);
        for (i = 0; i < BLOCK_NUM; i += 4) {
            for (j = 0; j < 4; j++) {
                temp[j] = (i + j + block * BLOCK_NUM) % PIC_WIDTH;
                temp[j + 4] = (i + j + block * BLOCK_NUM) / PIC_WIDTH;
            }
            xx = (v_float){temp[0], temp[1], temp[2], temp[3]};
            yy = (v_float){temp[4], temp[5], temp[6], temp[7]};
            a = spu_madd(xstep, xx, int2vfloat(rc.left));
            b = spu_madd(ystep, yy, int2vfloat(rc.top));
            x1 = int2vfloat(0);
            y1 = int2vfloat(0);
            over = spu_splats((uint) 0);
            cnt = spu_splats((uint) 1);
            for (j = 1; j < MAX_CALC; j++) {
                x2 = spu_sub(spu_madd(x1, x1, a), spu_mul(y1, y1));
                y2 = spu_madd(spu_mul(x1, y1), int2vfloat(2), b);
                r = spu_madd(x2, x2, spu_mul(y2, y2));
                over = spu_or(over, spu_cmpgt(r, int2vfloat(4)));
                cnt = spu_sel(spu_splats(j + 1), cnt, over);
                chk = spu_and(over, spu_rlqwbyte(over, 8));
                chk = spu_and(chk, spu_rlqwbyte(chk, 4));
                if (spu_extract(chk, 0)) break;
                x1 = spu_sel(x2, int2vfloat(0), over);
                y1 = spu_sel(y2, int2vfloat(0), over);
            }
            cnt = spu_sel(cnt, spu_splats((uint) 0),
                          spu_cmpeq(cnt, spu_splats((uint) MAX_CALC)));
            *memp = cnt;
            memp++;
            if (i == BLOCK_NUM / 2)
                send_mem(0, block);
        }
        send_mem(1, block);
#else
        // without SIMD
        xstep = rc.width / PIC_WIDTH;
        ystep = rc.height / PIC_HEIGHT;
        for (i = 0; i < BLOCK_NUM; i++) {
            xx = (i + block * BLOCK_NUM) % PIC_WIDTH;
            yy = (i + block * BLOCK_NUM) / PIC_WIDTH;
            a = rc.left + xstep * xx;
            b = rc.top + ystep * yy;
            x1 = 0;
            y1 = 0;
            for (j = 1; j < MAX_CALC; j++) {
                x2 = x1 * x1 - y1 * y1 + a;
                y2 = x1 * y1 * 2 + b;
                r = x2 * x2 + y2 * y2;
                if (r > 4) break;
                x1 = x2;
                y1 = y2;
            }
            mem[i] = j & (-(j != MAX_CALC));
            if (i == BLOCK_NUM / 2)
                send_mem(0, block);
        }
        send_mem(1, block);
#endif
    }
    wait_dma(1);
    return 0;
}

int main()
{
    spe_initialize();
    calc();
    return 0;
}
#endif /* _SPESOURCE_ */
