/* ppsort2 -- partial sort program / libspe2 version   |
|   2007.12.10 Naoki FUJIEDA, Arch Lab. TOKYO TECH    */
#include "define.h"
#include <unistd.h>

#ifdef _PPESOURCE_

#define TIMEBASE 79800000.0f
#define MSEC(x) (x / TIMEBASE * 1.0e3)

uint *sort_data;

void initialize(spe_context_ptr_t *spe)
{
    int i;
    FILE *fp;

    // memory allocate
    if (posix_memalign((void **)&sort_data, DMA_ALIGN, SORT_MEM)) {
        fprintf(stderr,"error: memalign failed\n");
        exit(1);
    }
    // read file
    if ((fp = fopen("rand.dat", "rb")) == NULL) {
        fprintf(stderr, "file open error\n");
        exit(1);
    }
    if ((fread(sort_data, sizeof(uint), SORT_NUM, fp)) != SORT_NUM) {
        fprintf(stderr, "file read error\n");
        exit(1);
    }
    fclose(fp);

    for (i = 0; i < SPE_NUM; i++) {
        dmapack[i].spe_num = SPE_NUM;
        dmapack[i].spe_id  = i + 1;
        /* set ppe2spe[] */
        dmapack[i].ppe2spe[0] = (uint) sort_data;
    }

#ifdef MEMIMAGE
    printf("Writing to %s. Wait a moment...", MEMIMAGE);
    fflush(stdout);
    int *p;
    if ((fp = fopen(MEMIMAGE, "wb")) == NULL) {
        fprintf(stderr, "file can't open.\n");
        return;
    }
    fprintf(fp, "SimCell_PPE_MEMORY_IMAGE_FILE_VERSION_1.0\n");
    
    p = (int *)dmapack;
    for (i = 0; i < sizeof(dmapack) / 4; i++)
        fprintf(fp, "@mem %08x %08x\n", (uint) dmapack + 4 * i, p[i]);

    for (i = 0; i < SORT_NUM; i++)
        fprintf(fp, "@mem %08x %08x\n",
                (uint) sort_data + 4 * i, sort_data[i]);

    for (i = 0; i < SPE_NUM; i++)
        fprintf(fp, "@mal %08x %08x\n", i + 1, (uint) &dmapack[i]);
    
    fprintf(fp, "@end %08x %08x\n", 0, 0);
    fclose(fp);
    printf("Done.\n");
#endif /* MEMIMAGE */
}

void finalize()
{
    int i, max;
    max = 0;
    for (i = 0; i < SPE_NUM; i++)
        if (max < dmapack[i].spe2ppe[0])
            max = dmapack[i].spe2ppe[0];

    printf("*** spu_time %9.2f [msec]\n", MSEC(max));
    free(sort_data);
}

#endif /* _PPESOURCE_ */

/**********************************************************************/

#ifdef _SPESOURCE_

#define INIT_VAL 0x7fffffffu

volatile uint buf[BLOCK_NUM] __attribute__ ((aligned(DMA_ALIGN)));

inline void wait_dma(int dma_tag)
{
    spu_writech(MFC_WrTagMask, 1 << dma_tag);
    spu_mfcstat(MFC_TAG_UPDATE_ALL);
}

void sort_stage1(uint *mem, int left, int right) 
{
    int i, j;
    uint pivot, temp;

    pivot = mem[(left + right) / 2];
    mem[(left + right) / 2] = mem[left];
    j = left;
    for (i = left + 1; i <= right; i++) {
        if (mem[i] < pivot) {
            j++;
            temp = mem[i];
            mem[i] = mem[j];
            mem[j] = temp;
        }
    }
    mem[left] = mem[j];
    mem[j] = pivot;
    if (j - left > 1)
        sort_stage1(mem, left, j - 1);
    if (right - j > 1)
        sort_stage1(mem, j + 1, right);
}

int main()
{
    int i;
    int spe_num, spe_id;
    uint mem_addr;

    spe_initialize();
    spe_num = dmapack.spe_num;
    spe_id = dmapack.spe_id;
    mem_addr = dmapack.ppe2spe[0];
    spu_write_decrementer(INIT_VAL);
    
    for (i = spe_id; i < TOTAL_BLOCK; i += spe_num) {
        mfc_get(buf, mem_addr + i * BLOCK_MEM, BLOCK_MEM, DMA_TAG0, 0, 0);
        wait_dma(DMA_TAG0);
        
        sort_stage1((uint *)buf, 0, (BLOCK_NUM - 1));
        
        mfc_put(buf, mem_addr + i * BLOCK_MEM, BLOCK_MEM, DMA_TAG0, 0, 0);
        wait_dma(DMA_TAG0);
    }

    dmapack.spe2ppe[0] = INIT_VAL - spu_read_decrementer();
    mfc_put(&dmapack, ea_dmapack, sizeof(dmapack), DMA_TAG0, 0, 0);
    wait_dma(DMA_TAG0);
    return 0;
}

#endif /* _SPESOURCE_ */
